#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import uuid

from gbpservice.contrib.nfp.config_orchestrator.common import common
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import (
    firewall)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import (
    loadbalancer)
from gbpservice.contrib.nfp.config_orchestrator.handlers.config import vpn
from gbpservice.contrib.nfp.config_orchestrator.handlers.notification import (
    handler as notif_handler)
from gbpservice.nfp.lib import transport

from neutron import context as ctx
from neutron.tests import base


class TestContext(object):

    def get_context(self):
        try:
            return ctx.Context('some_user', 'some_tenant')
        except Exception:
            return ctx.Context('some_user', 'some_tenant')


class Conf(object):

    class Test_RPC(object):

        def __init__(self):
            self.topic = 'xyz_topic'

    def __init__(self):
        self.host = 'dummy_host'
        self.backend = 'rpc'
        self.RPC = self.Test_RPC()


class RpcMethods(object):

    def cast(self, context, method, **kwargs):
        return

    def call(self, context, method, **kwargs):
        return {}


def call_network_function_info():
    data = {'network_function': {
        'id': str(uuid.uuid4()),
        'description': {}
    }}
    return data


class GeneralConfigStructure(object):

    def _check_general_structure(self, request_data, rsrc_name, resource=None):
        flag = 0
        if all(key in request_data for key in ["info", "config"]):
            header_data = request_data['info']
            if all(key in header_data for key in ["context", "service_type",
                                                  "service_vendor"]):
                if not self._check_resource_header_data(rsrc_name,
                                                        header_data["context"],
                                                        resource):
                    return False
                data = request_data['config']
                for ele in data:
                    if all(key in ele for key in ["resource",
                                                  "resource_data"]):
                        if self._check_resource_structure(rsrc_name,
                                                          ele['resource_data'],
                                                          resource):
                            flag = 1
                        else:
                            flag = 0
                    else:
                        flag = 0
        if flag == 1:
            return True
        return False

    def verify_firewall_structure(self, blob_data, resource=None):
        if all(k in blob_data for k in ["neutron_context", "host",
                                        "firewall"]):
            context = blob_data['neutron_context']
            try:
                if context['service_info']:
                    data = context['service_info']
                    if all(k in data for k in ["firewalls",
                                               "firewall_policies",
                                               "firewall_rules"]):
                        return True
            except AttributeError:
                return False
        return False

    def verify_firewall_header_data(self, data, resource=None):
        if all(k in data for k in ["neutron_context", "network_function_id",
                                   "fw_mac", "requester"]):
            if data['requester'] == 'nas_service':
                return True
            return False

    def verify_loadbalancer_header_data(self, data, resource=None):
        if all(k in data for k in ["neutron_context", "requester"]):
            if resource == "vip":
                if not all(k in data for k in ["network_function_id",
                                               "vip_id"]):
                    return False
            if data['requester'] == 'nas_service':
                return True
        return False

    def verify_vpn_header_data(self, data, resource=None):
        if all(k in data for k in ["neutron_context", "requester"]):
            if resource == "ipsec_site_connection":
                if not all(k in data for k in ["network_function_id",
                                               "ipsec_site_connection_id"]):
                    return False
            if data['requester'] == 'nas_service':
                return True
            return False

    def verify_loadbalancer_structure(self, blob_data, resource):
        if all(k in blob_data for k in ["neutron_context", resource]):
            context = blob_data["neutron_context"]
            try:
                if context['service_info']:
                    data = context['service_info']
                    if all(k in data for k in ["pools", "vips", "members",
                                               "health_monitors"]):
                        return True
            except AttributeError:
                return False
        return False

    def verify_vpn_structure(self, blob_data, resource):
        if all(k in blob_data for k in ["neutron_context", "resource",
                                        "rsrc_id", "reason"]):
            context = blob_data["neutron_context"]
            try:
                if context['service_info']:
                    data = context['service_info']
                    if resource.lower() == "vpn_service":
                        if all(k in data for k in ["vpnservices"]):
                            return True
                    elif resource.lower() == "ipsec_site_connection":
                        if all(k in data for k in ["vpnservices",
                                                   "ikepolicies",
                                                   "ipsecpolicies",
                                                   "ipsec_site_conns"]):
                            return True
            except AttributeError:
                return False
        return False

    def _check_resource_structure(self, rsrc_name, data, resource=None):
        mod = self
        mod_method = getattr(mod, "verify_%s_structure" % rsrc_name)
        return mod_method(data, resource)

    def _check_resource_header_data(self, rsrc_name, data, resource):
        mod = self
        mod_method = getattr(mod, "verify_%s_header_data" % rsrc_name)
        return mod_method(data, resource)


class FirewallTestCase(base.BaseTestCase):

    def setUp(self):
        super(FirewallTestCase, self).setUp()
        self.conf = Conf()
        self.fw_handler = firewall.FwAgent(self.conf, 'sc')
        self.context = TestContext().get_context()
        self.rpc_methods = RpcMethods()
        self.fw = self._firewall_data()
        self.host = 'host'
        import_path = ("neutron_fwaas.db.firewall.firewall_db."
                       "Firewall_db_mixin")
        self.import_fw_api = import_path + '.get_firewalls'
        self.import_fwp_api = import_path + '.get_firewall_policies'
        self.import_fwr_api = import_path + '.get_firewall_rules'
        self.import_lib = 'gbpservice.nfp.lib.transport'
        self._call = 'oslo_messaging.rpc.client._CallContext.call'

    def _firewall_data(self):
        return {'tenant_id': str(uuid.uuid4()),
                'description': str({'network_function_id': str(uuid.uuid4())}),
                'firewall_policy_id': str(uuid.uuid4())
                }

    def _cast_firewall(self, conf, context, body,
                       method_type, device_config=False,
                       network_function_event=False):
        g_cnfg = GeneralConfigStructure()
        self.assertTrue(g_cnfg._check_general_structure(body, 'firewall'))

    def _call_to_get_network_function_desc(self):
        data = call_network_function_info()
        data['network_function']['description'] = "\n" + str(
            {'provider_ptg_info': [str(uuid.uuid4())],
             'service_vendor': 'xyz'})
        return data['network_function']

    def test_create_firewall(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_fw_api) as gfw, mock.patch(
                self.import_fwp_api) as gfwp, mock.patch(
                self.import_fwr_api) as gfwr, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gfw.return_value = []
            gfwp.return_value = []
            gfwr.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)

            mock_call.side_effect = self._call_to_get_network_function_desc
            mock_send.side_effect = self._cast_firewall
            self.fw_handler.create_firewall(self.context, self.fw, self.host)

    def test_delete_firewall(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_fw_api) as gfw, mock.patch(
                self.import_fwp_api) as gfwp, mock.patch(
                self.import_fwr_api) as gfwr, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gfw.return_value = []
            gfwp.return_value = []
            gfwr.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)

            mock_call.side_effect = self._call_to_get_network_function_desc
            mock_send.side_effect = self._cast_firewall
            self.fw_handler.delete_firewall(self.context, self.fw, self.host)


class LoadBalanceTestCase(base.BaseTestCase):

    def setUp(self):
        super(LoadBalanceTestCase, self).setUp()
        self.conf = Conf()
        self.lb_handler = loadbalancer.LbAgent(self.conf, 'sc')
        self.context = TestContext().get_context()
        import_path = ("neutron_lbaas.db.loadbalancer.loadbalancer_db."
                       "LoadBalancerPluginDb")
        self.import_gp_api = import_path + '.get_pools'
        self.import_gv_api = import_path + '.get_vips'
        self.import_gm_api = import_path + '.get_members'
        self.import_ghm_api = import_path + '.get_health_monitors'
        self.import_lib = 'gbpservice.nfp.lib.transport'
        self._call = 'oslo_messaging.rpc.client._CallContext.call'
        self._get_pool = import_path + '.get_pool'

    def _cast_loadbalancer(self, conf, context, body,
                           method_type, device_config=False,
                           network_function_event=False):
        g_cnfg = GeneralConfigStructure()
        try:
            resource = body['config'][0]['resource']
            if resource == 'pool_health_monitor':
                resource = 'health_monitor'
            self.assertTrue(g_cnfg._check_general_structure(
                body, 'loadbalancer', resource))
        except Exception:
            self.assertTrue(False)

    def _call_to_get_network_function_desc(self):
        data = call_network_function_info()
        data['network_function']['description'] = ("\n" + str(
            {'service_vendor': 'xyz'}))
        return data['network_function']

    def _call_data(self, context, method, **kwargs):
        if method.lower() == "get_network_function_details":
            data = call_network_function_info()
            data['network_function']['description'] = "\n" + str(
                {'service_vendor': 'xyz'})
            return data['network_function']

        return []

    def _loadbalancer_data(self, resource):
        data = {'tenant_id': str(uuid.uuid4()),
                'id': str(uuid.uuid4())
                }
        if resource.lower() not in ['member', 'health_monitor']:
            desc = str({'network_function_id': str(uuid.uuid4())})
            data.update({'description': desc})
        if resource.lower() == 'vip':
            data.update({'pool_id': str(uuid.uuid4())})
        return data

    def _get_mocked_pool(self, context, pool_id):
        return {'id': pool_id,
                'description': str({'network_function_id': str(uuid.uuid4())})}

    def test_create_vip(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            vip = self._loadbalancer_data('vip')
            self.lb_handler.create_vip(self.context, vip)

    def test_update_vip(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            old_vip = self._loadbalancer_data('vip')
            vip = self._loadbalancer_data('vip')
            self.lb_handler.update_vip(self.context, old_vip, vip)

    def test_delete_vip(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            vip = self._loadbalancer_data('vip')
            self.lb_handler.delete_vip(self.context, vip)

    def test_create_pool(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            pool = self._loadbalancer_data('pool')
            driver_name = "dummy"
            self.lb_handler.create_pool(self.context, pool, driver_name)

    def test_update_pool(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            old_pool = self._loadbalancer_data('pool')
            pool = self._loadbalancer_data('pool')
            self.lb_handler.update_pool(self.context, old_pool, pool)

    def test_delete_pool(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            pool = self._loadbalancer_data('pool')
            self.lb_handler.delete_pool(self.context, pool)

    def test_create_member(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            member = self._loadbalancer_data('member')
            member.update({'pool_id': str(uuid.uuid4())})
            self.lb_handler.create_member(self.context, member)

    def test_update_member(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            old_member = self._loadbalancer_data('member')
            member = self._loadbalancer_data('member')
            pool_id = str(uuid.uuid4())
            old_member.update({'pool_id': pool_id})
            member.update({'pool_id': pool_id})
            self.lb_handler.update_member(self.context, old_member, member)

    def test_delete_member(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            member = self._loadbalancer_data('member')
            member.update({'pool_id': str(uuid.uuid4())})
            self.lb_handler.delete_member(self.context, member)

    def test_create_pool_health_monitor(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            hm = self._loadbalancer_data('health_monitor')
            pool_id = str(uuid.uuid4())
            self.lb_handler.create_pool_health_monitor(
                self.context, hm, pool_id)

    def test_update_pool_health_monitor(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)
            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            old_hm = self._loadbalancer_data('health_monitor')
            hm = self._loadbalancer_data('health_monitor')
            pool_id = str(uuid.uuid4())
            self.lb_handler.update_pool_health_monitor(
                self.context, old_hm, hm, pool_id)

    def test_delete_pool_health_monitor(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gp_api) as gp, mock.patch(
                self.import_gv_api) as gv, mock.patch(
                self.import_gm_api) as gm, mock.patch(
                self.import_ghm_api) as ghm, mock.patch(
                self._call) as mock_call, mock.patch(
                self._get_pool) as mock_pool, mock.patch(
                import_send) as mock_send:
            gp.return_value = []
            gv.return_value = []
            gm.return_value = []
            ghm.return_value = []
            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)

            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_loadbalancer
            mock_pool.side_effect = self._get_mocked_pool
            hm = self._loadbalancer_data('health_monitor')
            pool_id = str(uuid.uuid4())
            self.lb_handler.delete_pool_health_monitor(
                self.context, hm, pool_id)


class VPNTestCase(base.BaseTestCase):

    def setUp(self):
        super(VPNTestCase, self).setUp()
        self.conf = Conf()
        self.vpn_handler = vpn.VpnAgent(self.conf, 'sc')
        self.context = TestContext().get_context()
        import_path = "neutron_vpnaas.db.vpn.vpn_db.VPNPluginDb"
        self.import_gvs_api = import_path + '.get_vpnservices'
        self.import_gikp_api = import_path + '.get_ikepolicies'
        self.import_gipsp_api = import_path + '.get_ipsecpolicies'
        self.import_gisc_api = import_path + '.get_ipsec_site_connections'
        self.import_lib = 'gbpservice.nfp.lib.transport'
        self._call = 'oslo_messaging.rpc.client._CallContext.call'

    def _cast_vpn(self, conf, context, body,
                  method_type, device_config=False,
                  network_function_event=False):
        g_cnfg = GeneralConfigStructure()
        try:
            resource = body['config'][0]['resource']
            self.assertTrue(g_cnfg._check_general_structure(
                body, 'vpn', resource))
        except Exception:
            self.assertTrue(False)

    def _call_data(self, context, method, **kwargs):
        if method.lower() == "get_network_function_details":
            data = call_network_function_info()
            data['network_function']['description'] = ("\n" + (
                "ipsec_site_connection_id=%s;service_vendor=xyz" % (
                    str(uuid.uuid4()))))
            return data['network_function']

        return []

    def _prepare_request_data(self, reason, rsrc_type):
        resource = {'tenant_id': str(uuid.uuid4()),
                    'id': str(uuid.uuid4()),
                    'description': (
                        "{'network_function_id':'%s'}" % (str(uuid.uuid4())))
                    }
        if rsrc_type.lower() == 'ipsec_site_connection':
            resource.update({'vpnservice_id': str(uuid.uuid4()),
                             'ikepolicy_id': str(uuid.uuid4()),
                             'ipsecpolicy_id': str(uuid.uuid4())})
        elif rsrc_type.lower() == 'vpn_service':
            resource.update({'subnet_id': str(uuid.uuid4()),
                             'router_id': str(uuid.uuid4())})
        return {'resource': resource,
                'rsrc_type': rsrc_type,
                'reason': reason,
                'rsrc_id': str(uuid.uuid4())
                }

    def _call_to_get_network_function_desc(self):
        data = call_network_function_info()
        data['network_function']['description'] = ("\n" + (
            "ipsec_site_connection_id=%s;service_vendor=xyz" % (
                str(uuid.uuid4()))))
        return data['network_function']

    def _get_networks(self):
        return []

    def _get_routers(self):
        return []

    def test_update_vpnservice_for_vpnservice(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gvs_api) as gvs, mock.patch(
                self.import_gikp_api) as gikp, mock.patch(
                self.import_gipsp_api) as gipsp, mock.patch(
                self.import_gisc_api) as gisc, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gvs.return_value = []
            gikp.return_value = []
            gipsp.return_value = []
            gisc.return_value = []

            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)

            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_vpn
            rsrc_type = 'vpn_service'
            reason = 'create'
            kwargs = self._prepare_request_data(reason, rsrc_type)
            self.vpn_handler.vpnservice_updated(self.context, **kwargs)

    def test_update_vpnservice_for_ipsec_site_connection(self):
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(self.import_gvs_api) as gvs, mock.patch(
                self.import_gikp_api) as gikp, mock.patch(
                self.import_gipsp_api) as gipsp, mock.patch(
                self.import_gisc_api) as gisc, mock.patch(
                self._call) as mock_call, mock.patch(
                import_send) as mock_send:
            gvs.return_value = []
            gikp.return_value = []
            gipsp.return_value = []
            gisc.return_value = []

            network_function_desc = self._call_to_get_network_function_desc()
            common.get_network_function_details = mock.MagicMock(
                return_value=network_function_desc)

            networks = self._get_networks()
            routers = self._get_routers()
            common.get_networks = mock.MagicMock(return_value=networks)
            common.get_routers = mock.MagicMock(return_value=routers)

            mock_call.side_effect = self._call_data
            mock_send.side_effect = self._cast_vpn
            rsrc_type = 'ipsec_site_connection'
            reason = 'delete'
            kwargs = self._prepare_request_data(reason, rsrc_type)
            self.vpn_handler.vpnservice_updated(self.context, **kwargs)


class FirewallNotifierTestCase(base.BaseTestCase):

    class Controller(object):

        def new_event(self, **kwargs):
            return

        def post_event(self, event):
            return

    def setUp(self):
        super(FirewallNotifierTestCase, self).setUp()
        self.conf = Conf()
        self.n_handler = notif_handler.NaasNotificationHandler(
            self.conf, self.Controller())
        self.context = TestContext().get_context()

    def _get_rpc_client(self):
        class Context(object):
            def cast(self, context, method, host='', firewall_id='', body=''):
                return {}

        class RCPClient(object):
            def __init__(self):
                self.cctxt = Context()

        return RCPClient()

    def get_notification_data(self):
        return {'info': {'service_type': 'firewall',
                         'context': {'logging_context': {}}},
                'notification': [{'data': {'firewall_id': '123',
                                           'status': 'set_firewall_status',
                                           'notification_type':
                                                     'set_firewall_status',
                                           'host': 'localhost'}
                                  }]
                }

    def test_set_firewall_status(self):
        notification_data = self.get_notification_data()
        rpc_client = self._get_rpc_client()
        transport.RPCClient = mock.MagicMock(return_value=rpc_client)
        self.n_handler.handle_notification(self.context,
                                           notification_data)

    def test_set_firewall_deleted(self):
        notification_data = self.get_notification_data()
        notification_data['notification'][0]['data'][
                          'notification_type'] = 'firewall_deleted'
        rpc_client = self._get_rpc_client()
        transport.RPCClient = mock.MagicMock(return_value=rpc_client)
        self.n_handler.handle_notification(self.context,
                                           notification_data)


class LoadbalancerNotifierTestCase(base.BaseTestCase):

    class Controller(object):

        def new_event(self, **kwargs):
            return

        def post_event(self, event):
            return

    def setUp(self):
        super(LoadbalancerNotifierTestCase, self).setUp()
        self.conf = Conf()
        self.n_handler = notif_handler.NaasNotificationHandler(
            self.conf, self.Controller())
        self.context = TestContext().get_context()

    def _get_rpc_client(self):
        class Context(object):
            def cast(self, context, method, host='', pool_id='',
                     stats='', body=''):
                return {}

        class RCPClient(object):
            def __init__(self):
                self.cctxt = Context()

        return RCPClient()

    def get_notification_data(self):
        return {'info': {'service_type': 'loadbalancer',
                         'context': {'logging_context': {}}},
                'notification': [{'data': {'obj_type': 'lb',
                                           'obj_id': '123',
                                           'status': 'set_firewall_status',
                                           'notification_type':
                                                     'update_status',
                                           'host': 'localhost'}
                                  }]
                }

    def test_update_status(self):
        notification_data = self.get_notification_data()
        rpc_client = self._get_rpc_client()
        transport.RPCClient = mock.MagicMock(return_value=rpc_client)
        self.n_handler.handle_notification(self.context,
                                           notification_data)

    def test_update_pool_stats(self):
        notification_data = self.get_notification_data()
        notification_data['notification'][0]['data'][
                          'notification_type'] = 'update_pool_stats'
        rpc_client = self._get_rpc_client()
        transport.RPCClient = mock.MagicMock(return_value=rpc_client)
        self.n_handler.handle_notification(self.context,
                                           notification_data)


class VpnNotifierTestCase(base.BaseTestCase):

    class Controller(object):

        def new_event(self, **kwargs):
            return

        def post_event(self, event):
            return

    def setUp(self):
        super(VpnNotifierTestCase, self).setUp()
        self.conf = Conf()
        self.n_handler = notif_handler.NaasNotificationHandler(
            self.conf, self.Controller())
        self.context = TestContext().get_context()

    def _get_rpc_client(self):
        class Context(object):
            def cast(self, context, method, host='', status='', body=''):
                return {}

        class RCPClient(object):
            def __init__(self):
                self.cctxt = Context()

        return RCPClient()

    def get_notification_data(self):
        return {'info': {'service_type': 'vpn',
                         'context': {'logging_context': {}}},
                'notification': [{'data': {'firewall_id': '123',
                                           'status': 'set_firewall_status',
                                           'notification_type':
                                                     'update_status',
                                           'host': 'localhost'}
                                  }]
                }

    def test_update_status(self):
        notification_data = self.get_notification_data()
        rpc_client = self._get_rpc_client()
        transport.RPCClient = mock.MagicMock(return_value=rpc_client)
        self.n_handler.handle_notification(self.context,
                                           notification_data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from sqlalchemy.orm import exc

from gbpservice.nfp.common import exceptions as nfp_exc
from gbpservice.nfp.orchestrator.db import common_db_mixin
from gbpservice.nfp.orchestrator.db import nfp_db_model

from gbpservice.nfp.core import log as nfp_logging
from neutron._i18n import _LW

LOG = nfp_logging.getLogger(__name__)


class NFPDbBase(common_db_mixin.CommonDbMixin):

    def __init__(self, *args, **kwargs):
        super(NFPDbBase, self).__init__(*args, **kwargs)

    def create_network_function(self, session, network_function):
        with session.begin(subtransactions=True):
            network_function_db = nfp_db_model.NetworkFunction(
                id=uuidutils.generate_uuid(),
                name=network_function['name'],
                description=network_function.get('description'),
                tenant_id=network_function['tenant_id'],
                service_id=network_function['service_id'],
                service_chain_id=network_function.get('service_chain_id'),
                service_profile_id=network_function['service_profile_id'],
                service_config=network_function.get('service_config'),
                config_policy_id=network_function.get('config_policy_id'),
                status=network_function['status'])
            session.add(network_function_db)
            return self._make_network_function_dict(network_function_db)

    def _get_network_function(self, session, network_function_id):
        try:
            return self._get_by_id(
                session, nfp_db_model.NetworkFunction, network_function_id)
        except exc.NoResultFound:
            raise nfp_exc.NetworkFunctionNotFound(
                network_function_id=network_function_id)

    def update_network_function(self, session, network_function_id,
                                updated_network_function):
        with session.begin(subtransactions=True):
            network_function_db = self._get_network_function(
                session, network_function_id)
            network_function_db.update(updated_network_function)
        return self._make_network_function_dict(network_function_db)

    def delete_network_function(self, session, network_function_id):
        with session.begin(subtransactions=True):
            network_function_db = self._get_network_function(
                session, network_function_id)
            session.delete(network_function_db)

    def get_network_function(self, session, network_function_id, fields=None):
        service = self._get_network_function(session, network_function_id)
        return self._make_network_function_dict(service, fields)

    def get_network_functions(self, session, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        marker_obj = self._get_marker_obj(
            'network_functions', limit, marker)
        return self._get_collection(session, nfp_db_model.NetworkFunction,
                                    self._make_network_function_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def _set_port_info_for_nfi(self, session, network_function_instance_db,
                               network_function_instance, is_update=False):
        nfi_db = network_function_instance_db
        port_info = network_function_instance.get('port_info')
        with session.begin(subtransactions=True):
            nfi_db.port_info = []
            for port in port_info:
                port_info_db = nfp_db_model.PortInfo(
                    id=port['id'],
                    port_model=port['port_model'],
                    port_classification=port.get('port_classification'),
                    port_role=port.get('port_role'))
                if is_update:
                    session.merge(port_info_db)
                else:
                    session.add(port_info_db)
                session.flush()  # Any alternatives for flush ??
                assoc = nfp_db_model.NSIPortAssociation(
                    network_function_instance_id=(
                        network_function_instance_db['id']),
                    data_port_id=port['id'])
                nfi_db.port_info.append(assoc)
            del network_function_instance['port_info']

    def create_network_function_instance(self, session,
                                         network_function_instance):
        with session.begin(subtransactions=True):
            network_function_instance_db = (
                nfp_db_model.NetworkFunctionInstance(
                    id=uuidutils.generate_uuid(),
                    name=network_function_instance['name'],
                    tenant_id=network_function_instance['tenant_id'],
                    description=network_function_instance.get('description'),
                    network_function_id=network_function_instance[
                        'network_function_id'],
                    network_function_device_id=network_function_instance.get(
                        'network_function_device_id'),
                    ha_state=network_function_instance.get('ha_state'),
                    status=network_function_instance['status']))
            session.add(network_function_instance_db)
            self._set_port_info_for_nfi(session, network_function_instance_db,
                                        network_function_instance)
        return self._make_network_function_instance_dict(
            network_function_instance_db)

    def _get_network_function_instance(self, session,
                                       network_function_instance_id):
        try:
            return self._get_by_id(
                session,
                nfp_db_model.NetworkFunctionInstance,
                network_function_instance_id)
        except exc.NoResultFound:
            raise nfp_exc.NetworkFunctionInstanceNotFound(
                network_function_instance_id=network_function_instance_id)

    def update_network_function_instance(self, session,
                                         network_function_instance_id,
                                         updated_network_function_instance):
        with session.begin(subtransactions=True):
            network_function_instance_db = self._get_network_function_instance(
                session, network_function_instance_id)
            if 'port_info' in updated_network_function_instance:
                self._set_port_info_for_nfi(
                    session,
                    network_function_instance_db,
                    updated_network_function_instance, is_update=True)
            network_function_instance_db.update(
                updated_network_function_instance)
        return self._make_network_function_instance_dict(
            network_function_instance_db)

    def delete_network_function_instance(self, session,
                                         network_function_instance_id):
        with session.begin(subtransactions=True):
            network_function_instance_db = self._get_network_function_instance(
                session, network_function_instance_id)
            for port in network_function_instance_db.port_info:
                self.delete_port_info(session, port['data_port_id'])
            session.delete(network_function_instance_db)

    def get_network_function_instance(self, session,
                                      network_function_instance_id,
                                      fields=None):
        network_function_instance = self._get_network_function_instance(
            session, network_function_instance_id)
        return self._make_network_function_instance_dict(
            network_function_instance, fields)

    def get_network_function_instances(self, session, filters=None,
                                       fields=None, sorts=None, limit=None,
                                       marker=None, page_reverse=False):
        port_info = None
        marker_obj = self._get_marker_obj(
            'network_function_instances', limit, marker)
        if filters:
            port_info = filters.pop('port_info', None)
        nfis = self._get_collection(
            session, nfp_db_model.NetworkFunctionInstance,
            self._make_network_function_instance_dict,
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker_obj=marker_obj, page_reverse=page_reverse)
        filtered_nfis = []
        if port_info:
            for nfi in nfis:
                if port_info == nfi['port_info']:
                    filtered_nfis.append(nfi)
            return filtered_nfis
        return nfis

    def _set_mgmt_port_for_nfd(self, session, network_function_device_db,
                               network_function_device, is_update=False):
        nfd_db = network_function_device_db
        mgmt_port_id = network_function_device.get('mgmt_port_id')
        if not mgmt_port_id:
            nfd_db.mgmt_port_id = None
            return
        with session.begin(subtransactions=True):
            port_info_db = nfp_db_model.PortInfo(
                id=mgmt_port_id['id'],
                port_model=mgmt_port_id['port_model'],
                port_classification=mgmt_port_id['port_classification'],
                port_role=mgmt_port_id['port_role'])
            if is_update:
                session.merge(port_info_db)
            else:
                session.add(port_info_db)
            session.flush()
            nfd_db.mgmt_port_id = port_info_db['id']

    def _set_monitoring_port_id_for_nfd(self, session,
                                        network_function_device_db,
                                        network_function_device,
                                        is_update=False):
        nfd_db = network_function_device_db
        monitoring_port_id = network_function_device.get(
            'monitoring_port_id')
        if not monitoring_port_id:
            nfd_db.monitoring_port_id = None
            return
        with session.begin(subtransactions=True):
            port_info_db = nfp_db_model.PortInfo(
                id=monitoring_port_id['id'],
                port_model=monitoring_port_id['port_model'],
                port_classification=monitoring_port_id[
                    'port_classification'],
                port_role=monitoring_port_id['port_role'])
            if is_update:
                session.merge(port_info_db)
            else:
                session.add(port_info_db)
            session.flush()
            nfd_db.monitoring_port_id = monitoring_port_id['id']

    def _set_monitoring_port_network_for_nfd(self, session,
                                             network_function_device_db,
                                             network_function_device,
                                             is_update=False):
        nfd_db = network_function_device_db
        monitoring_port_network = network_function_device.get(
            'monitoring_port_network')
        if not monitoring_port_network:
            nfd_db.monitoring_port_network = None
            return
        with session.begin(subtransactions=True):
            network_info_db = nfp_db_model.NetworkInfo(
                id=monitoring_port_network['id'],
                network_model=monitoring_port_network['network_model'])
            session.add(network_info_db)
            session.flush()
            nfd_db.monitoring_port_network = (
                monitoring_port_network['id'])
            del network_function_device['monitoring_port_network']

    def _set_provider_metadata_for_nfd(self, session,
                                       network_function_device_db,
                                       network_function_device,
                                       is_update=False):
        nfd_db = network_function_device_db
        provider_metadata = nfd_db['provider_metadata']

        if is_update:
            if provider_metadata:
                provider_metadata = jsonutils.loads(provider_metadata)
            updated_provider_metadata_str = network_function_device.pop(
                'provider_metadata', {})
            if not updated_provider_metadata_str:
                return
            if updated_provider_metadata_str:
                updated_provider_metadata = jsonutils.loads(
                    updated_provider_metadata_str)
            if (type(updated_provider_metadata) is dict and
                    updated_provider_metadata and provider_metadata):
                updated_provider_metadata.update(provider_metadata)
            provider_metadata_str = jsonutils.dumps(updated_provider_metadata)
        else:
            if not provider_metadata:
                provider_metadata_str = ''
                return
            provider_metadata_str = jsonutils.dumps(provider_metadata)
        nfd_db.provider_metadata = provider_metadata_str

    def create_network_function_device(self, session, network_function_device):
        with session.begin(subtransactions=True):
            network_function_device_db = nfp_db_model.NetworkFunctionDevice(
                id=(network_function_device.get('id') or
                    uuidutils.generate_uuid()),
                name=network_function_device['name'],
                description=network_function_device.get('description'),
                tenant_id=network_function_device['tenant_id'],
                mgmt_ip_address=network_function_device[
                    'mgmt_ip_address'],
                service_vendor=network_function_device.get('service_vendor'),
                max_interfaces=network_function_device['max_interfaces'],
                reference_count=network_function_device['reference_count'],
                interfaces_in_use=network_function_device['interfaces_in_use'],
                status=network_function_device['status'],
                gateway_port=network_function_device.get('gateway_port'))
            session.add(network_function_device_db)
            self._set_mgmt_port_for_nfd(
                session, network_function_device_db, network_function_device)
            self._set_monitoring_port_id_for_nfd(
                session, network_function_device_db, network_function_device)
            self._set_monitoring_port_network_for_nfd(
                session, network_function_device_db, network_function_device)
            self._set_provider_metadata_for_nfd(
                session, network_function_device_db, network_function_device)
            return self._make_network_function_device_dict(
                network_function_device_db)

    def _get_network_function_device(self, session,
                                     network_function_device_id):
        try:
            nfd = self._get_by_id(
                session,
                nfp_db_model.NetworkFunctionDevice,
                network_function_device_id)
            return nfd
        except exc.NoResultFound:
            raise nfp_exc.NetworkFunctionDeviceNotFound(
                network_function_device_id=network_function_device_id)

    def update_network_function_device(self, session,
                                       network_function_device_id,
                                       updated_network_function_device):
        with session.begin(subtransactions=True):
            network_function_device_db = self._get_network_function_device(
                session, network_function_device_id)
            if updated_network_function_device.get('provider_metadata'):
                updated_network_function_device[
                    'provider_metadata'] = jsonutils.dumps(
                    updated_network_function_device[
                        'provider_metadata'])
            if updated_network_function_device.get('mgmt_port_id'):
                self._set_mgmt_port_for_nfd(
                    session,
                    network_function_device_db,
                    updated_network_function_device,
                    is_update=True)

            if 'monitoring_port_id' in updated_network_function_device:
                self._set_monitoring_port_id_for_nfd(
                    session,
                    network_function_device_db,
                    updated_network_function_device,
                    is_update=True)
            if 'monitoring_port_network' in updated_network_function_device:
                self._set_monitoring_port_network_for_nfd(
                    session,
                    network_function_device_db,
                    updated_network_function_device,
                    is_update=True)
            self._set_provider_metadata_for_nfd(
                session, network_function_device_db,
                updated_network_function_device,
                is_update=True)
            mgmt_port_id = (
                updated_network_function_device.pop('mgmt_port_id', None))
            if mgmt_port_id:
                updated_network_function_device[
                    'mgmt_port_id'] = mgmt_port_id['id']

            monitoring_port_id = (
                updated_network_function_device.pop('monitoring_port_id',
                                                    None))
            if monitoring_port_id:
                updated_network_function_device[
                    'monitoring_port_id'] = monitoring_port_id['id']
            network_function_device_db.update(updated_network_function_device)
            updated_network_function_device['mgmt_port_id'] = mgmt_port_id
            updated_network_function_device[
                'monitoring_port_id'] = monitoring_port_id

            return self._make_network_function_device_dict(
                network_function_device_db)

    def delete_network_function_device(self, session,
                                       network_function_device_id):
        with session.begin(subtransactions=True):
            network_function_device_db = self._get_network_function_device(
                session, network_function_device_id)
            if network_function_device_db.mgmt_port_id:
                self.delete_port_info(session,
                                      network_function_device_db.mgmt_port_id)
            if network_function_device_db.monitoring_port_id:
                self.delete_port_info(
                    session,
                    network_function_device_db.monitoring_port_id)
            if network_function_device_db.monitoring_port_network:
                self.delete_network_info(
                    session,
                    network_function_device_db.monitoring_port_network)
            session.delete(network_function_device_db)

    def get_network_function_device(self, session, network_function_device_id,
                                    fields=None):
        network_function_device = self._get_network_function_device(
            session, network_function_device_id)
        return self._make_network_function_device_dict(
            network_function_device, fields)

    def get_network_function_devices(self, session, filters=None, fields=None,
                                     sorts=None, limit=None, marker=None,
                                     page_reverse=False):
        marker_obj = self._get_marker_obj(
            'network_function_devices', limit, marker)
        return self._get_collection(session,
                                    nfp_db_model.NetworkFunctionDevice,
                                    self._make_network_function_device_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def increment_network_function_device_count(self, session,
                                                network_function_device_id,
                                                field_name,
                                                updated_value=1):
        with session.begin(subtransactions=True):
            network_function_device = self._get_network_function_device(
                session, network_function_device_id)
            value = network_function_device[field_name]
            value += updated_value
            update_device = (
                {field_name: value})
            self.update_network_function_device(session,
                                                network_function_device_id,
                                                update_device)

    def decrement_network_function_device_count(self, session,
                                                network_function_device_id,
                                                field_name,
                                                updated_value=1):
        with session.begin(subtransactions=True):
            network_function_device = self._get_network_function_device(
                session, network_function_device_id)
            value = network_function_device[field_name]
            value -= updated_value
            update_device = (
                {field_name: value})
            self.update_network_function_device(session,
                                                network_function_device_id,
                                                update_device)

    def get_port_info(self, session, port_id, fields=None):
        port_info = self._get_port_info(session, port_id)
        return self._make_port_info_dict(port_info, fields)

    def _get_port_info(self, session, port_id):
        try:
            return self._get_by_id(
                session, nfp_db_model.PortInfo, port_id)
        except exc.NoResultFound:
            raise nfp_exc.NFPPortNotFound(port_id=port_id)

    def delete_port_info(self, session, port_id):
        with session.begin(subtransactions=True):
            port_info_db = self._get_port_info(session, port_id)
            session.delete(port_info_db)

    def delete_network_info(self, session, network_id):
        with session.begin(subtransactions=True):
            network_info_db = self._get_network_info(session, network_id)
            session.delete(network_info_db)

    def get_network_info(self, session, network_id, fields=None):
        network_info = self._get_network_info(session, network_id)
        return self._make_network_info_dict(network_info, fields)

    def _get_network_info(self, session, network_id):
        return self._get_by_id(
            session, nfp_db_model.NetworkInfo, network_id)

    def _make_port_info_dict(self, port_info, fields):
        res = {
            'id': port_info['id'],
            'port_classification': port_info['port_classification'],
            'port_model': port_info['port_model'],
            'port_role': port_info['port_role']
        }
        return res

    def _make_network_info_dict(self, network_info, fields):
        res = {
            'id': network_info['id'],
            'network_model': network_info['network_model'],
        }
        return res

    def _make_network_function_dict(self, network_function, fields=None):
        res = {'id': network_function['id'],
               'tenant_id': network_function['tenant_id'],
               'name': network_function['name'],
               'description': network_function['description'],
               'service_id': network_function['service_id'],
               'service_chain_id': network_function['service_chain_id'],
               'service_profile_id': network_function['service_profile_id'],
               'service_config': network_function['service_config'],
               'config_policy_id': network_function['config_policy_id'],
               'status': network_function['status']
               }
        res['network_function_instances'] = [
            nfi['id'] for nfi in network_function[
                'network_function_instances']]
        return res

    def _make_network_function_instance_dict(self, nfi, fields=None):
        res = {'id': nfi['id'],
               'tenant_id': nfi['tenant_id'],
               'name': nfi['name'],
               'description': nfi['description'],
               'ha_state': nfi['ha_state'],
               'network_function_id': nfi['network_function_id'],
               'network_function_device_id': nfi['network_function_device_id'],
               'status': nfi['status']
               }
        res['port_info'] = [
            port['data_port_id'] for port in nfi['port_info']]
        return res

    def _make_network_function_device_dict(self, nfd, fields=None):
        res = {'id': nfd['id'],
               'tenant_id': nfd['tenant_id'],
               'name': nfd['name'],
               'description': nfd['description'],
               'mgmt_ip_address': nfd['mgmt_ip_address'],
               'mgmt_port_id': nfd['mgmt_port_id'],
               'monitoring_port_id': nfd['monitoring_port_id'],
               'monitoring_port_network': nfd['monitoring_port_network'],
               'service_vendor': nfd['service_vendor'],
               'max_interfaces': nfd['max_interfaces'],
               'reference_count': nfd['reference_count'],
               'interfaces_in_use': nfd['interfaces_in_use'],
               'status': nfd['status'],
               'gateway_port': nfd.get('gateway_port')
               }
        if nfd.get('provider_metadata'):
            res.update({'provider_metadata': nfd['provider_metadata']})
        return res

    def add_cluster_info(self, session, cluster_info):
        with session.begin(subtransactions=True):
            cluster_info = nfp_db_model.ClusterInfo(
                id=cluster_info['id'], tenant_id=cluster_info['tenant_id'],
                network_function_device_id=cluster_info[
                    'network_function_device_id'],
                cluster_group=cluster_info[
                    'cluster_group'], virtual_ip=cluster_info['virtual_ip'],
                multicast_ip=cluster_info.get('multicast_ip', None),
                cluster_name=cluster_info.get('cluster_name', None)
            )
            session.add(cluster_info)
            return cluster_info

    def insert_cluster_records(self, session, cluster_infos):
        with session.begin(subtransactions=True):
            for cluster_info in cluster_infos:
                cluster_info = nfp_db_model.ClusterInfo(
                    id=cluster_info['id'],
                    tenant_id=cluster_info['tenant_id'],
                    network_function_device_id=cluster_info[
                        'network_function_device_id'],
                    cluster_group=cluster_info['cluster_group'],
                    virtual_ip=cluster_info['virtual_ip'],
                    multicast_ip=cluster_info.get('multicast_ip', None),
                    cluster_name=cluster_info.get('cluster_name', None))
                session.add(cluster_info)

    def get_cluster_info(self, session, _id):
        try:
            return self._get_by_id(
                session,
                nfp_db_model.ClusterInfo, _id)
        except exc.NoResultFound:
            raise nfp_exc.ClusterInfoNotFound(id=_id)

    def get_all_cluster_info(self, session, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        marker_obj = self._get_marker_obj(
            'nfd_cluster_mapping_info', limit, marker)
        return self._get_collection(session,
                                    nfp_db_model.ClusterInfo,
                                    self._get_cluster_info_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    def del_cluster_info(self, session, _id):
        with session.begin(subtransactions=True):
            cluster_info = self.get_cluster_info(session, _id)
            session.delete(cluster_info)

    def delete_cluster_info(self, session, port_id_list):
        for port_id in port_id_list:
            self.del_cluster_info(session, port_id)

    def _get_cluster_info_dict(self, cluster_info, filters=None, fields=None,
                               sorts=None, limit=None, marker=None,
                               page_reverse=False):
        return {
            'id': cluster_info['id'], 'tenant_id': cluster_info['tenant_id'],
            'network_function_device_id': cluster_info[
                'network_function_device_id'],
            'cluster_group': cluster_info['cluster_group'],
            'multicast_ip': cluster_info['multicast_ip'],
            'cluster_name': cluster_info['cluster_name']
        }

    def add_service_gateway_details(self, session, service_gw_details):
        primary_gw_vip_pt, secondary_gw_vip_pt = self._get_vip_pt_ids(
            service_gw_details.get('gateway_vips'))
        if isinstance(service_gw_details['primary_instance_gw_pt'], dict):
            primary_instance_gw_pt = service_gw_details[
                'primary_instance_gw_pt']['id']
            secondary_instance_gw_pt = service_gw_details.get(
                'secondary_instance_gw_pt', {}).get('id')
        else:
            primary_instance_gw_pt = service_gw_details[
                'primary_instance_gw_pt']
            secondary_instance_gw_pt = service_gw_details.get(
                'secondary_instance_gw_pt')
        with session.begin(subtransactions=True):
            gw_detail = nfp_db_model.ServiceGatewayDetails(
                id=service_gw_details['id'],
                network_function_id=service_gw_details[
                    'network_function_id'],
                gateway_ptg=service_gw_details['gw_ptg'],
                primary_instance_gw_pt=primary_instance_gw_pt,
                secondary_instance_gw_pt=secondary_instance_gw_pt,
                primary_gw_vip_pt=primary_gw_vip_pt,
                secondary_gw_vip_pt=secondary_gw_vip_pt
            )
            session.add(gw_detail)
            return gw_detail

    def _get_vip_pt_ids(self, vips):
        if not vips:
            return None, None
        else:
            if isinstance(vips, list):
                primary_gw_vip_pt = vips[0]['id']
                secondary_gw_vip_pt = vips[1]['id'] if len(vips) == 2 else None
                return primary_gw_vip_pt, secondary_gw_vip_pt
            elif isinstance(vips, dict):
                return vips['primary_gw_vip_pt'], vips['secondary_gw_vip_pt']
            else:
                return None, None

    def get_providers_for_gateway(self, session, _id):
        svc_gw = nfp_db_model.ServiceGatewayDetails
        try:
            with session.begin(subtransactions=True):
                return self._get_gw_info_dict(session.query(svc_gw).filter(
                    svc_gw.gateway_ptg == _id).all())
        except exc.NoResultFound:
            raise

    def get_gateway_detail(self, session, nf_id):
        svc_gw = nfp_db_model.ServiceGatewayDetails
        try:
            with session.begin(subtransactions=True):
                return self._get_gw_info_dict(session.query(svc_gw).filter(
                    svc_gw.network_function_id == nf_id).one())
        except exc.NoResultFound:
            LOG.warning(_LW("Gateway detail doesn't exist for Network Function"
                            " %s ") % nf_id)
            raise

    def _get_gw_info_dict(self, gw):
        if not gw:
            return
        if isinstance(gw, list):
            return [dict(id=info['id'],
                         network_function_id=info['network_function_id'],
                         gateway_ptg=info['gateway_ptg'],
                         primary_instance_gw_pt=info['primary_instance_gw_pt'],
                         secondary_instance_gw_pt=info[
                             'secondary_instance_gw_pt'],
                         primary_gw_vip_pt=info['primary_gw_vip_pt'],
                         secondary_gw_vip_pt=info['secondary_gw_vip_pt']
                         ) for info in gw]

        return {'id': gw['id'],
                'network_function_id': gw['network_function_id'],
                'gateway_ptg': gw['gateway_ptg'],
                'primary_instance_gw_pt': gw['primary_instance_gw_pt'],
                'secondary_instance_gw_pt': gw['secondary_instance_gw_pt'],
                'primary_gw_vip_pt': gw['primary_gw_vip_pt'],
                'secondary_gw_vip_pt': gw['secondary_gw_vip_pt']}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

FW_NFP_CONFIGAGENT_TOPIC = 'nfp-firewall-agent'
LB_NFP_CONFIGAGENT_TOPIC = 'nfp-lbaas-agent'
LBV2_NFP_CONFIGAGENT_TOPIC = 'nfp-lbaasv2-agent'
VPN_NFP_CONFIGAGENT_TOPIC = 'nfp-vpn_agent'
NFP_NSO_TOPIC = "nfp-service-orchestrator"

FW_NFP_PLUGIN_TOPIC = 'q-firewall-plugin'
LB_NFP_PLUGIN_TOPIC = 'n-lbaas-plugin'
LBV2_NFP_PLUGIN_TOPIC = 'n-lbaasv2-plugin'
VPN_NFP_PLUGIN_TOPIC = 'vpn_plugin'
DEVICE_ORCH_TOPIC = 'nfp-configurator-ndo'
SERVICE_ORCH_TOPIC = 'nfp-configurator-nso'
CONFIG_ORCH_TOPIC = 'nfp-nco-notification-topic'



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webob.exc

import mock
from neutron.common import config  # noqa
from neutron.common import exceptions as n_exc
from neutron import context as n_context
from neutron.db import api as db_api
from neutron.db import model_base
from neutron import manager
from neutron.plugins.common import constants as pconst
from oslo_config import cfg
from oslo_serialization import jsonutils

from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db  # noqa
from gbpservice.neutron.services.grouppolicy import config as gpconfig  # noqa
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    context as ncp_context)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    exceptions as exc)
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    plugin as ncp_plugin)
import gbpservice.neutron.services.servicechain.plugins.ncp.config  # noqa
from gbpservice.neutron.services.servicechain.plugins.ncp.node_drivers import (
    dummy_driver as dummy_driver)
from gbpservice.neutron.tests.unit.db.grouppolicy import test_group_policy_db
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_gp_driver)
from gbpservice.neutron.tests.unit.services.servicechain import (
    test_servicechain_plugin as test_base)


class ServiceChainNCPTestPlugin(ncp_plugin.NodeCompositionPlugin):

    supported_extension_aliases = ['servicechain'] + (
        test_group_policy_db.UNSUPPORTED_REQUIRED_EXTS)
    path_prefix = "/servicechain"


SC_PLUGIN_KLASS = (ServiceChainNCPTestPlugin.__module__ + '.' +
                   ServiceChainNCPTestPlugin.__name__)
CORE_PLUGIN = test_gp_driver.CORE_PLUGIN
GP_PLUGIN_KLASS = (
    "gbpservice.neutron.services.grouppolicy.plugin.GroupPolicyPlugin"
)
CHAIN_TENANT_ID = 'sci_owner'


class NodeCompositionPluginTestMixin(object):

    DEFAULT_LB_CONFIG = '{}'
    SERVICE_PROFILE_VENDOR = 'dummy'

    @property
    def sc_plugin(self):
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get(pconst.SERVICECHAIN)
        return servicechain_plugin

    def _create_service_profile(self, **kwargs):
        """Create service profile wrapper that can be used by drivers."""
        return self.create_service_profile(**kwargs)

    def _create_redirect_rule(self, spec_id):
        action = self.create_policy_action(action_type='REDIRECT',
                                           action_value=spec_id)
        classifier = self.create_policy_classifier(
            port_range=80, protocol='tcp', direction='bi')
        rule = self.create_policy_rule(
            policy_actions=[action['policy_action']['id']],
            policy_classifier_id=classifier['policy_classifier']['id'])
        return rule

    def _create_redirect_prs(self, spec_id):
        rule = self._create_redirect_rule(spec_id)['policy_rule']
        prs = self.create_policy_rule_set(policy_rules=[rule['id']])
        return prs

    def _create_simple_service_chain(self, number_of_nodes=1,
                                     service_type='LOADBALANCER'):
        prof = self.create_service_profile(
            service_type=service_type,
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']

        node_ids = []
        for x in xrange(number_of_nodes):
            node_ids.append(self.create_servicechain_node(
                service_profile_id=prof['id'],
                config=self.DEFAULT_LB_CONFIG,
                expected_res_status=201)['servicechain_node']['id'])

        return self._create_chain_with_nodes(node_ids)

    def _create_chain_with_nodes(self, node_ids=None):
        node_ids = node_ids or []
        spec = self.create_servicechain_spec(
            nodes=node_ids,
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        provider = self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})['policy_target_group']
        consumer = self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})['policy_target_group']
        return provider, consumer, prs

    def _add_node_driver(self, name):
        inst = dummy_driver.NoopNodeDriver()
        inst.initialize(name)
        ext = mock.Mock()
        ext.obj = inst
        self.sc_plugin.driver_manager.ordered_drivers.append(ext)
        self.sc_plugin.driver_manager.drivers[name] = ext


class NodeCompositionPluginTestCase(
        test_base.TestGroupPolicyPluginGroupResources,
        NodeCompositionPluginTestMixin):

    def setUp(self, core_plugin=None, gp_plugin=None, node_drivers=None,
              node_plumber=None):
        cfg.CONF.set_override(
            'extension_drivers', ['proxy_group'], group='group_policy')
        if node_drivers:
            cfg.CONF.set_override('node_drivers', node_drivers,
                                  group='node_composition_plugin')
        cfg.CONF.set_override('node_plumber', node_plumber or 'dummy_plumber',
                              group='node_composition_plugin')
        config.cfg.CONF.set_override('policy_drivers',
                                     ['implicit_policy', 'resource_mapping',
                                      'chain_mapping'],
                                     group='group_policy')
        super(NodeCompositionPluginTestCase, self).setUp(
            core_plugin=core_plugin or CORE_PLUGIN,
            gp_plugin=gp_plugin or GP_PLUGIN_KLASS,
            sc_plugin=SC_PLUGIN_KLASS)
        engine = db_api.get_engine()
        model_base.BASEV2.metadata.create_all(engine)
        self.driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj

    def _create_simple_chain(self):
        node = self._create_profiled_servicechain_node(
            service_type="LOADBALANCER",
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']])['servicechain_spec']

        action = self.create_policy_action(
            action_type='REDIRECT', action_value=spec['id'])['policy_action']
        classifier = self.create_policy_classifier(
            direction='bi', port_range=80, protocol='tcp')['policy_classifier']
        rule = self.create_policy_rule(
            policy_classifier_id=classifier['id'],
            policy_actions=[action['id']])['policy_rule']

        prs = self.create_policy_rule_set(
            policy_rules=[rule['id']])['policy_rule_set']

        provider = self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})['policy_target_group']
        consumer = self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})['policy_target_group']

        return provider, consumer, node

    def test_spec_ordering_list_servicechain_instances(self):
        pass

    def test_context_attributes(self):
        # Verify Context attributes for simple config
        plugin_context = n_context.get_admin_context()
        profile = self._create_service_profile(
            service_type="LOADBALANCER",
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=profile['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']])['servicechain_spec']
        provider = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target_group()
        management = self.create_policy_target_group(
            service_management=True,
            is_admin_context=True)['policy_target_group']
        classifier = self.create_policy_classifier()['policy_classifier']
        instance = self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']], classifier_id=classifier['id'])[
                                                    'servicechain_instance']

        # Verify created without errors
        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node)

        self.assertIsNotNone(ctx.gbp_plugin)
        self.assertIsNotNone(ctx.sc_plugin)
        self.assertIsNotNone(ctx.core_plugin)
        self.assertIsNotNone(ctx.plugin_context)
        self.assertIsNotNone(ctx.plugin_session)
        self.assertIsNotNone(ctx.session)
        self.assertIsNotNone(ctx.admin_context)
        self.assertIsNotNone(ctx.admin_session)
        del ctx.current_profile['nodes']
        self.assertEqual(ctx.current_profile, profile)
        self.assertEqual(instance['id'], ctx.instance['id'])
        self.assertEqual(provider['id'], ctx.provider['id'])
        self.assertIsNone(ctx.consumer)
        self.assertEqual(management['id'], ctx.management['id'])
        self.assertEqual([spec['id']], [x['id'] for x in ctx.relevant_specs])
        self.assertIsNone(ctx.original_node)
        self.assertEqual(0, len(ctx.get_service_targets()))

        instance['provider_ptg_id'] = 'dummy-id'
        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node)
        self.assertIsNone(ctx.provider)
        self.assertIsNone(ctx.consumer)

    def test_context_relevant_specs(self):
        plugin_context = n_context.get_admin_context()
        node_used = self._create_profiled_servicechain_node(
            service_type="LOADBALANCER",
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        spec_used = self.create_servicechain_spec(
            nodes=[node_used['id']])['servicechain_spec']

        provider = self.create_policy_target_group()['policy_target_group']
        classifier = self.create_policy_classifier()['policy_classifier']
        instance = self.create_servicechain_instance(
            provider_ptg_id=provider['id'],
            classifier_id=classifier['id'],
            servicechain_specs=[spec_used['id']])['servicechain_instance']

        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node_used)
        self.assertEqual([spec_used['id']],
                         [x['id'] for x in ctx.relevant_specs])

    def test_manager_initialized(self):
        mgr = self.plugin.driver_manager
        self.assertIsInstance(mgr.ordered_drivers[0].obj,
                              dummy_driver.NoopNodeDriver)
        for driver in mgr.ordered_drivers:
            self.assertTrue(driver.obj.initialized)

    def test_spec_parameters(self):
        """Test that config_param_names is empty when using NCP.
        In NCP the config attribute of a node may be something different than
        a HEAT template, therefore config_param_names is not used.
        """

        params_node_1 = ['p1', 'p2', 'p3']
        params_node_2 = ['p4', 'p5', 'p6']
        params_node_3 = ['p7', 'p8', 'p9']

        def params_dict(params):
            return jsonutils.dumps({'Parameters':
                                    dict((x, {}) for x in params)})

        prof = self._create_service_profile(
            service_type='LOADBALANCER', shared=True,
            vendor=self.SERVICE_PROFILE_VENDOR,
            tenant_id='admin')['service_profile']

        # Create 2 nodes with different parameters
        node1 = self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            config=params_dict(params_node_1),
            expected_res_status=201)['servicechain_node']
        node2 = self.create_servicechain_node(
            service_profile_id=prof['id'], shared=True,
            config=params_dict(params_node_2),
            expected_res_status=201)['servicechain_node']

        # Create SC spec with the nodes assigned
        spec = self.create_servicechain_spec(
            nodes=[node1['id'], node2['id']], shared=True,
            expected_res_status=201)['servicechain_spec']

        # Verify param names is empty
        self.assertIsNone(spec['config_param_names'])

        # Update the spec removing one node
        self.update_servicechain_spec(spec['id'], nodes=[node1['id']],
                                      expected_res_status=200)

        spec = self.show_servicechain_spec(spec['id'])['servicechain_spec']
        # Verify param names is empty
        self.assertIsNone(spec['config_param_names'])

        # Update a node with new config params
        self.update_servicechain_node(node1['id'],
                                      config=params_dict(params_node_3),
                                      expected_res_status=200)

        spec = self.show_servicechain_spec(spec['id'])['servicechain_spec']
        # Verify param names is empty
        self.assertIsNone(spec['config_param_names'])

    def test_create_service_chain(self):
        deploy = self.driver.create = mock.Mock()
        destroy = self.driver.delete = mock.Mock()

        self._create_simple_service_chain(1)
        self.assertEqual(1, deploy.call_count)
        self.assertEqual(0, destroy.call_count)

        deploy.reset_mock()

        provider, _, _ = self._create_simple_service_chain(3)
        self.assertEqual(3, deploy.call_count)
        self.assertEqual(0, destroy.call_count)

        self.update_policy_target_group(provider['id'],
                                        provided_policy_rule_sets={})
        self.assertEqual(3, deploy.call_count)
        self.assertEqual(3, destroy.call_count)

    def test_update_service_chain(self):
        deploy = self.driver.create = mock.Mock()
        update = self.driver.update = mock.Mock()
        destroy = self.driver.delete = mock.Mock()

        provider, _, prs = self._create_simple_service_chain(1)
        self.assertEqual(1, deploy.call_count)
        self.assertEqual(0, destroy.call_count)

        # REVISIT(Magesh): When bug #1446587 is fixed, we should test by
        # performing a classifier or rule update instead of SC instance update
        instances = self._list('servicechain_instances')[
                                            'servicechain_instances']
        self.assertEqual(1, len(instances))
        self.update_servicechain_instance(
            instances[0]['id'],
            expected_res_status=200)
        self.assertEqual(1, update.call_count)
        self.assertEqual(0, destroy.call_count)

    def test_create_service_chain_fails(self):
        deploy = self.driver.create = mock.Mock()
        destroy = self.driver.delete = mock.Mock()

        deploy.side_effect = Exception

        try:
            self._create_simple_service_chain(3)
        except Exception:
            pass

        self.assertEqual(1, deploy.call_count)
        self.assertEqual(3, destroy.call_count)

    def test_update_node_fails(self):
        validate_update = self.driver.validate_update = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']

        node_id = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']['id']

        spec = self.create_servicechain_spec(
            nodes=[node_id],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        validate_update.side_effect = exc.NodeCompositionPluginBadRequest(
            resource='node', msg='reason')

        res = self.update_servicechain_node(node_id,
                                            description='somethingelse',
                                            expected_res_status=400)
        self.assertEqual('NodeCompositionPluginBadRequest',
                         res['NeutronError']['type'])

    def test_update_instantiated_profile_fails(self):
        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']

        node_id = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']['id']

        spec = self.create_servicechain_spec(
            nodes=[node_id], expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        res = self.update_service_profile(prof['id'],
                                          vendor='somethingelse',
                                          expected_res_status=400)
        self.assertEqual('ServiceProfileInUseByAnInstance',
                         res['NeutronError']['type'])

    def test_second_driver_scheduled_if_first_fails(self):
        self._add_node_driver('test')
        drivers = [x.obj for x in
                   self.sc_plugin.driver_manager.ordered_drivers]
        create_1 = drivers[0].validate_create = mock.Mock()
        create_1.side_effect = n_exc.NeutronException()

        # This happens without error
        profile = self._create_service_profile(
            service_type="TYPE",
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=profile['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']])['servicechain_spec']
        provider = self.create_policy_target_group()['policy_target_group']
        classifier = self.create_policy_classifier()['policy_classifier']
        self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']], classifier_id=classifier['id'],
            expected_res_status=201)

    def test_chain_fails_if_no_drivers_available(self):
        self._add_node_driver('test')
        drivers = [x.obj for x in
                   self.sc_plugin.driver_manager.ordered_drivers]
        create_1 = drivers[0].validate_create = mock.Mock()
        create_1.side_effect = n_exc.NeutronException()
        create_2 = drivers[1].validate_create = mock.Mock()
        create_2.side_effect = n_exc.NeutronException()

        profile = self._create_service_profile(
            service_type="TYPE",
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=profile['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']])['servicechain_spec']
        provider = self.create_policy_target_group()['policy_target_group']
        classifier = self.create_policy_classifier()['policy_classifier']
        self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']], classifier_id=classifier['id'],
            expected_res_status=400)

    def test_multiple_nodes_update(self):
        update = self.driver.update = mock.Mock()
        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']

        self._create_chain_with_nodes([node['id']])
        self.update_servicechain_node(node['id'], name='somethingelse')
        self.assertEqual(1, update.call_count)

        update.reset_mock()
        self._create_chain_with_nodes([node['id']])
        self._create_chain_with_nodes([node['id']])
        self.update_servicechain_node(node['id'], name='somethingelse')
        self.assertEqual(3, update.call_count)

    def test_inuse_spec_node_update_rejected(self):
        prof = self.create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']

        node1 = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        node2 = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node1['id'], node2['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        res = self.update_servicechain_spec(spec['id'],
                                            nodes=[node1['id']],
                                            expected_res_status=400)
        self.assertEqual('InuseSpecNodeUpdateNotAllowed',
                         res['NeutronError']['type'])

    def test_instance_update(self):
        prof = self.create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']

        node1 = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']
        node2 = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node1['id'], node2['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        instances = self._list('servicechain_instances')[
                                            'servicechain_instances']
        self.assertEqual(1, len(instances))
        spec2 = self.create_servicechain_spec(
            nodes=[node1['id']],
            expected_res_status=201)['servicechain_spec']
        res = self.update_servicechain_instance(
            instances[0]['id'], servicechain_specs=[spec2['id']],
            expected_res_status=200)
        self.assertEqual([spec2['id']],
                         res['servicechain_instance']['servicechain_specs'])

    def test_relevant_ptg_update(self):
        add = self.driver.update_policy_target_added = mock.Mock()
        rem = self.driver.update_policy_target_removed = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        provider = self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})['policy_target_group']
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        # Verify notification issued for created PT in the provider
        pt = self.create_policy_target(
            policy_target_group_id=provider['id'])['policy_target']
        pt['port_attributes'] = {}
        self.assertEqual(1, add.call_count)
        add.assert_called_with(mock.ANY, pt)
        del pt['port_attributes']

        # Verify notification issued for deleted PT in the provider
        self.delete_policy_target(pt['id'])
        self.assertEqual(1, rem.call_count)
        rem.assert_called_with(mock.ANY, pt)

    def test_irrelevant_ptg_update(self):
        add = self.driver.update_policy_target_added = mock.Mock()
        rem = self.driver.update_policy_target_removed = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node['id']], expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})

        other = self.create_policy_target_group()['policy_target_group']

        # Verify notification issued for created PT in the provider
        pt = self.create_policy_target(
            policy_target_group_id=other['id'])['policy_target']
        self.assertFalse(add.called)

        # Verify notification issued for deleted PT in the provider
        self.delete_policy_target(pt['id'])
        self.assertFalse(rem.called)

    def test_notify_chain_update_hook(self):
        update_hook = self.driver.notify_chain_parameters_updated = mock.Mock()

        prof = self.create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=201)['servicechain_spec']

        action = self.create_policy_action(action_type='REDIRECT',
                                           action_value=spec['id'])
        classifier = self.create_policy_classifier(
            port_range=80, protocol='tcp', direction='bi')['policy_classifier']
        rule = self.create_policy_rule(
            policy_actions=[action['policy_action']['id']],
            policy_classifier_id=classifier['id'])['policy_rule']
        prs = self.create_policy_rule_set(
            policy_rules=[rule['id']])['policy_rule_set']

        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})
        instances = self._list('servicechain_instances')[
            'servicechain_instances']
        self.assertEqual(1, len(instances))

        self.update_policy_classifier(classifier['id'], port_range=22)
        update_hook.assert_called_with(mock.ANY)

    def test_context_no_management(self):
        # Verify Context attributes for simple config
        plugin_context = n_context.get_admin_context()
        plugin_context.is_admin = False
        plugin_context.is_advsvc = False
        plugin_context.tenant_id = 'test-tenant'
        node = self._create_profiled_servicechain_node()['servicechain_node']
        spec = self.create_servicechain_spec(
            nodes=[node['id']])['servicechain_spec']
        provider = self.create_policy_target_group()['policy_target_group']

        # Verify admin created SM is None
        management = self.create_policy_target_group(
            service_management=True, tenant_id='admin',
            is_admin_context=True)['policy_target_group']
        pc = self.create_policy_classifier()['policy_classifier']
        instance = self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']],
            classifier_id=pc['id'])['servicechain_instance']
        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node)

        self.assertIsNone(ctx.management)

        self.delete_policy_target_group(management['id'],
                                        is_admin_context=True)
        shared_management = self.create_policy_target_group(
            service_management=True, tenant_id='admin',
            is_admin_context=True, shared=True)['policy_target_group']
        instance = self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']],
            classifier_id=pc['id'])['servicechain_instance']
        # Now admin Service Management PTG is visible
        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node)
        self.assertEqual(shared_management['id'], ctx.management['id'])

        # Private management overrides shared one
        private_management = self.create_policy_target_group(
            service_management=True,
            is_admin_context=True)['policy_target_group']
        instance = self.create_servicechain_instance(
            provider_ptg_id=provider['id'], consumer_ptg_id='N/A',
            servicechain_specs=[spec['id']],
            classifier_id=pc['id'])['servicechain_instance']
        ctx = ncp_context.get_node_driver_context(
            self.plugin, plugin_context, instance, node)
        self.assertEqual(private_management['id'], ctx.management['id'])

    def test_node_drivers_notified_consumer_event(self):
        add = self.driver.update_node_consumer_ptg_added = mock.Mock()
        rem = self.driver.update_node_consumer_ptg_removed = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        consumer = self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})['policy_target_group']

        # Verify notification issued for PTG consuming
        add.assert_called_with(mock.ANY, consumer)

        # Verify notification issued for PTG unconsuming
        consumer = self.update_policy_target_group(
            consumer['id'],
            consumed_policy_rule_sets={})['policy_target_group']
        rem.assert_called_with(mock.ANY, consumer)

        provider, consumer, prs = self._create_simple_service_chain(3)
        with mock.patch.object(ncp_plugin.NodeCompositionPlugin,
                "update_chains_consumer_removed") as ptg_removed:
                plugin_context = n_context.get_admin_context()
                self._gbp_plugin.delete_policy_target_group(plugin_context,
                        consumer['id'])

                self.assertEqual(ptg_removed.call_count, 1)
                consumer['consumed_policy_rule_sets'] = []
                ptg_removed.assert_called_once_with(
                        mock.ANY, consumer, mock.ANY)

        add.reset_mock()
        rem.reset_mock()

    def test_no_unrelated_chains_notified(self):
        add = self.driver.update_node_consumer_ptg_added = mock.Mock()
        rem = self.driver.update_node_consumer_ptg_removed = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        # This creates a chain
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})

        # Create a PRS and assign a consumer with no provider (hence, no chain)
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        ptg = self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})['policy_target_group']

        # No notification should be issued
        self.assertFalse(add.called)
        self.assertFalse(rem.called)

        # Remove the consumer
        self.update_policy_target_group(ptg['id'],
                                        consumed_policy_rule_sets={},
                                        expected_res_status=200)
        # No notification should be issued
        self.assertFalse(add.called)
        self.assertFalse(rem.called)

    def test_node_drivers_notified_provider_updated(self):
        upd = self.driver.policy_target_group_updated = mock.Mock()

        prof = self._create_service_profile(
            service_type='LOADBALANCER',
            vendor=self.SERVICE_PROFILE_VENDOR)['service_profile']
        node = self.create_servicechain_node(
            service_profile_id=prof['id'],
            config=self.DEFAULT_LB_CONFIG,
            expected_res_status=201)['servicechain_node']

        spec = self.create_servicechain_spec(
            nodes=[node['id']],
            expected_res_status=201)['servicechain_spec']
        prs = self._create_redirect_prs(spec['id'])['policy_rule_set']
        provider = self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})['policy_target_group']

        # Verify notification issued for PTG consuming
        upd.assert_called_with(mock.ANY, None, provider)
        upd.reset_mock()

        # Verify notification issued for PTG consuming
        new_provider = self.update_policy_target_group(
            provider['id'],
            consumed_policy_rule_sets={prs['id']: ''})['policy_target_group']
        upd.assert_called_with(mock.ANY, provider, new_provider)

        upd.reset_mock()


class TestQuotasForServiceChain(test_base.ServiceChainPluginTestCase):

    @property
    def sc_plugin(self):
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get(pconst.SERVICECHAIN)
        return servicechain_plugin

    def setUp(self, core_plugin=None, gp_plugin=None, node_drivers=None,
              node_plumber=None):
        if node_drivers:
            cfg.CONF.set_override('node_drivers', node_drivers,
                                  group='node_composition_plugin')
        cfg.CONF.set_override('node_plumber', node_plumber or 'dummy_plumber',
                              group='node_composition_plugin')
        config.cfg.CONF.set_override('policy_drivers',
                                     ['implicit_policy', 'resource_mapping',
                                      'chain_mapping'],
                                     group='group_policy')
        super(TestQuotasForServiceChain, self).setUp(
            core_plugin=core_plugin or CORE_PLUGIN,
            gp_plugin=gp_plugin or GP_PLUGIN_KLASS,
            sc_plugin=SC_PLUGIN_KLASS)
        engine = db_api.get_engine()
        model_base.BASEV2.metadata.create_all(engine)
        self.driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj
        cfg.CONF.set_override('quota_servicechain_node', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_spec', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_instance', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_service_profile', 1,
                              group='QUOTAS')

    def tearDown(self):
        cfg.CONF.set_override('quota_servicechain_node', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_spec', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_instance', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_service_profile', -1,
                              group='QUOTAS')
        super(TestQuotasForServiceChain, self).tearDown()

    def test_servicechain_node_quota(self):
        self.create_servicechain_node()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_node)

    def test_servicechain_spec_quota(self):
        self.create_servicechain_spec()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_spec)

    def test_servicechain_instance_quota(self):
        self.create_servicechain_instance()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_instance)

    def test_service_profile(self):
        self.create_service_profile(service_type=pconst.FIREWALL)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_service_profile,
                          service_type=pconst.FIREWALL)

    def test_quota_implicit_service_instance(self):
        prof = self.create_service_profile(
            service_type='LOADBALANCER',
            vendor="vendor")['service_profile']

        node1_id = self.create_servicechain_node(
            service_profile_id=prof['id'], config="{}",
            expected_res_status=201)['servicechain_node']['id']

        spec = self.create_servicechain_spec(
            nodes=[node1_id],
            expected_res_status=201)['servicechain_spec']
        action = self.create_policy_action(action_type='REDIRECT',
                                           action_value=spec['id'])
        classifier = self.create_policy_classifier(
            port_range=80, protocol='tcp', direction='bi')
        rule = self.create_policy_rule(
            policy_actions=[action['policy_action']['id']],
            policy_classifier_id=classifier['policy_classifier']['id'])
        prs = self.create_policy_rule_set(
            policy_rules=[rule['policy_rule']['id']])['policy_rule_set']
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: ''})
        self.create_policy_target_group(
            consumed_policy_rule_sets={prs['id']: ''})
        # Second service instance creation should fail now
        # sice service instance quota is 1, resulting in PTG
        # creation error
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target_group,
                          provided_policy_rule_sets={prs['id']: ''})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneclient import exceptions as k_exceptions
from keystoneclient.v2_0 import client as keyclient
from neutron._i18n import _LE
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice.common import utils
from gbpservice.neutron.services.servicechain.plugins.ncp.node_plumbers import(
    traffic_stitching_plumber as tscp)

LOG = logging.getLogger(__name__)

TSCP_OPTS = [
    cfg.StrOpt('plumbing_resource_owner_user',
           help=_("Username of the Openstack keystone user who owns the "
                  "resources created by the traffic stitching plumber")),
    cfg.StrOpt('plumbing_resource_owner_password',
               help=_("Openstack keystone password for the user who "
                      "owns the resources created by the traffic stitching "
                      "plumber"),
               secret=True),
    cfg.StrOpt('plumbing_resource_owner_tenant_name',
               help=_("Name of the Tenant that will own the plumber created "
                      " resources"),)
]

cfg.CONF.register_opts(TSCP_OPTS, "admin_owned_resources_apic_tscp")


class AdminOwnedResourcesApicTSCP(tscp.TrafficStitchingPlumber):
    """Traffic Stitching Plumber for APIC with Admin owned resources.

    This plumber for APIC mapping provides the ability to choose the user and
    who owns the resources created by the plumber.
    """

    def initialize(self):
        self._resource_owner_tenant_id = None
        super(AdminOwnedResourcesApicTSCP, self).initialize()

    @property
    def resource_owner_tenant_id(self):
        if not self._resource_owner_tenant_id:
            self._resource_owner_tenant_id = (
                self._get_resource_owner_tenant_id())
        return self._resource_owner_tenant_id

    def plug_services(self, context, deployment):
        context = self._get_resource_owner_context(context)
        super(AdminOwnedResourcesApicTSCP, self).plug_services(
            context, deployment)

    def unplug_services(self, context, deployment):
        context = self._get_resource_owner_context(context)
        super(AdminOwnedResourcesApicTSCP, self).unplug_services(
            context, deployment)

    def _get_resource_owner_tenant_id(self):
        user, pwd, tenant, auth_url = utils.get_keystone_creds()
        keystoneclient = keyclient.Client(username=user, password=pwd,
                                          auth_url=auth_url)
        try:
            tenant = keystoneclient.tenants.find(name=tenant)
            return tenant.id
        except k_exceptions.NotFound:
            with excutils.save_and_reraise_exception(reraise=True):
                LOG.error(_LE('No tenant with name %s exists.'), tenant)
        except k_exceptions.NoUniqueMatch:
            with excutils.save_and_reraise_exception(reraise=True):
                LOG.error(_LE('Multiple tenants matches found for %s'), tenant)

    def _get_resource_owner_context(self, context):
        resource_owner_context = context.elevated()
        resource_owner_context.tenant_id = self.resource_owner_tenant_id
        user, pwd, _, auth_url = utils.get_keystone_creds()
        keystoneclient = keyclient.Client(username=user, password=pwd,
                                          auth_url=auth_url)
        resource_owner_context.auth_token = keystoneclient.get_token(
            self.resource_owner_tenant_id)
        return resource_owner_context



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

SERVICE_TYPE = 'loadbalancer'
NEUTRON = 'neutron'

LBAAS_AGENT_RPC_TOPIC = 'lbaas_agent'
LBAAS_GENERIC_CONFIG_RPC_TOPIC = 'lbaas_generic_config'
LBAAS_PLUGIN_RPC_TOPIC = 'n-lbaas-plugin'
AGENT_TYPE_LOADBALANCER = 'OC Loadbalancer agent'

# Service operation status constants
ACTIVE = "ACTIVE"
DOWN = "DOWN"
CREATED = "CREATED"
PENDING_CREATE = "PENDING_CREATE"
PENDING_UPDATE = "PENDING_UPDATE"
PENDING_DELETE = "PENDING_DELETE"
INACTIVE = "INACTIVE"
ERROR = "ERROR"
STATUS_SUCCESS = "SUCCESS"

ACTIVE_PENDING_STATUSES = (
    ACTIVE,
    PENDING_CREATE,
    PENDING_UPDATE
)

""" HTTP request/response """
HAPROXY_AGENT_LISTEN_PORT = 1234
REQUEST_URL = "http://%s:%s/%s"
HTTP_REQ_METHOD_POST = 'POST'
HTTP_REQ_METHOD_GET = 'GET'
HTTP_REQ_METHOD_PUT = 'PUT'
HTTP_REQ_METHOD_DELETE = 'DELETE'
CONTENT_TYPE_HEADER = 'Content-type'
JSON_CONTENT_TYPE = 'application/json'

LB_METHOD_ROUND_ROBIN = 'ROUND_ROBIN'
LB_METHOD_LEAST_CONNECTIONS = 'LEAST_CONNECTIONS'
LB_METHOD_SOURCE_IP = 'SOURCE_IP'

PROTOCOL_TCP = 'TCP'
PROTOCOL_HTTP = 'HTTP'
PROTOCOL_HTTPS = 'HTTPS'

HEALTH_MONITOR_PING = 'PING'
HEALTH_MONITOR_TCP = 'TCP'
HEALTH_MONITOR_HTTP = 'HTTP'
HEALTH_MONITOR_HTTPS = 'HTTPS'

LBAAS = 'lbaas'

PROTOCOL_MAP = {
    PROTOCOL_TCP: 'tcp',
    PROTOCOL_HTTP: 'http',
    PROTOCOL_HTTPS: 'https',
}
BALANCE_MAP = {
    LB_METHOD_ROUND_ROBIN: 'roundrobin',
    LB_METHOD_LEAST_CONNECTIONS: 'leastconn',
    LB_METHOD_SOURCE_IP: 'source'
}
REQUEST_RETRIES = 0
REQUEST_TIMEOUT = 120

# Operations
CREATE = 'create'
UPDATE = 'update'
DELETE = 'delete'

""" Event ids """
EVENT_CREATE_POOL = 'CREATE_POOL'
EVENT_UPDATE_POOL = 'UPDATE_POOL'
EVENT_DELETE_POOL = 'DELETE_POOL'

EVENT_CREATE_VIP = 'CREATE_VIP'
EVENT_UPDATE_VIP = 'UPDATE_VIP'
EVENT_DELETE_VIP = 'DELETE_VIP'

EVENT_CREATE_MEMBER = 'CREATE_MEMBER'
EVENT_UPDATE_MEMBER = 'UPDATE_MEMBER'
EVENT_DELETE_MEMBER = 'DELETE_MEMBER'

EVENT_CREATE_POOL_HEALTH_MONITOR = 'CREATE_POOL_HEALTH_MONITOR'
EVENT_UPDATE_POOL_HEALTH_MONITOR = 'UPDATE_POOL_HEALTH_MONITOR'
EVENT_DELETE_POOL_HEALTH_MONITOR = 'DELETE_POOL_HEALTH_MONITOR'

EVENT_AGENT_UPDATED = 'AGENT_UPDATED'
EVENT_COLLECT_STATS = 'COLLECT_STATS'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.configurator.lib import demuxer
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                        fw_test_data as fo)
from neutron.tests import base


class ServiceAgentDemuxerTestCase(base.BaseTestCase):
    """ Implements test cases for demuxer of configurator. """
    def __init__(self, *args, **kwargs):
        super(ServiceAgentDemuxerTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()
        self.demuxer = demuxer.ServiceAgentDemuxer()
        self.maxDiff = None

    def test_get_service_type(self):
        """ Tests that demuxer extracts the service type from
        request data.

        Returns: none

        """

        expected_val = 'firewall'

        request_data = self.fo.fake_request_data_generic_single(self)
        actual_val = self.demuxer.get_service_type(request_data)

        self.assertEqual(actual_val, expected_val)

    def test_get_service_agent_info_generic_config(self):
        """ Tests that demuxer extracts and prepares a list of configurations
        of generic config for each request inside the request_data.

        Returns: none

        """

        expected_val = self.fo.fake_sa_req_list()

        request_data = self.fo.fake_request_data_generic_bulk()
        actual_val, service_type = self.demuxer.get_service_agent_info(
                                'create', 'firewall', request_data, True)

        for agent_info in expected_val:
            agent_info['agent_info']['context']['nf_id'] = 'nf_id'
            agent_info['agent_info']['context']['nfi_id'] = 'nfi_id'
        self.assertEqual(actual_val, expected_val)

    def test_get_service_agent_info_firewall(self):
        """ Tests that demuxer extracts and prepares a list of configuration
        of firewall for each request inside the request_data.

        Returns: none

        """

        expected_val = self.fo.fake_sa_req_list_fw()

        request_data = self.fo.fake_request_data_fw()
        actual_val, service_type = self.demuxer.get_service_agent_info(
                                'create', 'firewall', request_data, False)

        self.assertEqual(actual_val, expected_val)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import traceback

from gbpservice.contrib.nfp.config_orchestrator.common import (
    lbv2_constants as lbv2_const)
from gbpservice.contrib.nfp.config_orchestrator.common import (
    topics as a_topics)
from gbpservice.nfp.common import constants as const
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport

from neutron._i18n import _LI
import oslo_messaging as messaging


LOG = nfp_logging.getLogger(__name__)


class RpcHandler(object):
    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, conf, sc):
        super(RpcHandler, self).__init__()
        self.conf = conf
        self.sc = sc

    def network_function_notification(self, context, notification_data):
        module_context.init()
        try:
            LOG.info(_LI("Received NETWORK FUNCTION NOTIFICATION:"
                         "%(notification)s"),
                     {'notification': notification_data['notification']})
            if notification_data['info']['service_type'] is not None:
                handler = NaasNotificationHandler(self.conf, self.sc)
                handler.\
                    handle_notification(context, notification_data)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            msg = ("Generic exception (%s) while handling message (%s) : %s"
                   % (e,
                      notification_data,
                      traceback.format_exception(exc_type,
                                                 exc_value,
                                                 exc_traceback)))
            LOG.error(msg)


class FirewallNotifier(object):

    def __init__(self, conf, sc):
        self._sc = sc
        self._conf = conf

    def set_firewall_status(self, context, notification_data):
        nfp_context = module_context.init()
        notification = notification_data['notification'][0]

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        resource_data = notification['data']
        firewall_id = resource_data['firewall_id']
        status = resource_data['status']

        LOG.info(_LI("Received firewall configuration create complete API, "
                     "making an RPC call set firewall status for "
                     "firewall:%(firewall)s and status: %(status)s"),
                 {'firewall': firewall_id,
                  'status': status})

        # RPC call to plugin to set firewall status
        rpcClient = transport.RPCClient(a_topics.FW_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt.cast(context, 'set_firewall_status',
                             host=resource_data['host'],
                             firewall_id=firewall_id,
                             status=status)

    def firewall_deleted(self, context, notification_data):
        nfp_context = module_context.init()
        notification = notification_data['notification'][0]

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        resource_data = notification['data']
        firewall_id = resource_data['firewall_id']

        LOG.info(_LI("Received firewall_configuration_delete_complete API, "
                     "making an RPC call firewall_deleted  for firewall:"
                     "%(firewall)s "),
                 {'firewall': firewall_id})

        # RPC call to plugin to update firewall deleted
        rpcClient = transport.RPCClient(a_topics.FW_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt.cast(context, 'firewall_deleted',
                             host=resource_data['host'],
                             firewall_id=firewall_id)


class LoadbalancerNotifier(object):

    def __init__(self, conf, sc):
        self._sc = sc
        self._conf = conf

    def update_status(self, context, notification_data):
        nfp_context = module_context.init()
        notification = notification_data['notification'][0]

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        resource_data = notification['data']
        obj_type = resource_data['obj_type']
        obj_id = resource_data['obj_id']
        status = resource_data['status']

        LOG.info(_LI("Received LB's update_status API. Making an "
                     "update_status RPC call to plugin for %(obj_type)s:"
                     "%(obj_id)s with status:%(status)s"),
                 {'obj_type': obj_type,
                  'obj_id': obj_id,
                  'status': status})

        # RPC call to plugin to update status of the resource
        rpcClient = transport.RPCClient(a_topics.LB_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt = rpcClient.client.prepare(
            version=const.LOADBALANCER_RPC_API_VERSION)
        rpcClient.cctxt.cast(context, 'update_status',
                             obj_type=obj_type,
                             obj_id=obj_id,
                             status=status)

    def update_pool_stats(self, context, notification_data):
        nfp_context = module_context.init()
        notification = notification_data['notification'][0]

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        resource_data = notification['data']
        pool_id = resource_data['pool_id']
        stats = resource_data['stats']
        host = resource_data['host']

        LOG.info(_LI("Received LB's update_pool_stats API, making an "
                     "update_pool_stats RPC cast to plugin for updating "
                     "pool stats for pool: %(pool)s"),
                 {'pool': pool_id})

        # RPC cast to plugin to update stats of pool
        rpcClient = transport.RPCClient(a_topics.LB_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt = rpcClient.client.prepare(
            version=const.LOADBALANCER_RPC_API_VERSION)
        rpcClient.cctxt.cast(context, 'update_pool_stats',
                             pool_id=pool_id,
                             stats=stats,
                             host=host)

    def vip_deleted(self, context, notification_data):
        pass


class LoadbalancerV2Notifier(object):

    def __init__(self, conf, sc):
        self._sc = sc
        self._conf = conf

    def update_status(self, context, notification_data):
        nfp_context = module_context.init()
        notification = notification_data['notification'][0]

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        resource_data = notification['data']
        obj_type = resource_data['obj_type']
        obj_id = resource_data['obj_id']

        rpcClient = transport.RPCClient(a_topics.LBV2_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt = rpcClient.client.prepare(
            version=const.LOADBALANCERV2_RPC_API_VERSION)

        lb_p_status = const.ACTIVE
        lb_o_status = None
        obj_p_status = resource_data['provisioning_status']
        obj_o_status = resource_data['operating_status']

        LOG.info(_LI("Received LB's update_status API. Making an "
                     "update_status RPC call to plugin for %(obj_type)s:"
                     "%(obj_id)s with status: %(status)s"),
                 {'obj_type': obj_type,
                  'obj_id': obj_id,
                  'status': obj_p_status})

        if obj_type == 'healthmonitor':
            obj_o_status = None

        if obj_type != 'loadbalancer':
            rpcClient.cctxt.cast(context, 'update_status',
                                 obj_type=obj_type,
                                 obj_id=obj_id,
                                 provisioning_status=obj_p_status,
                                 operating_status=obj_o_status)
        else:
            lb_o_status = lbv2_const.ONLINE
            if obj_p_status == const.ERROR:
                lb_p_status = const.ERROR
                lb_o_status = lbv2_const.OFFLINE

        rpcClient.cctxt.cast(context, 'update_status',
                             obj_type='loadbalancer',
                             obj_id=resource_data['root_lb_id'],
                             provisioning_status=lb_p_status,
                             operating_status=lb_o_status)

    # TODO(jiahao): implememnt later
    def update_loadbalancer_stats(self, context, loadbalancer_id, stats_data):
        pass


class VpnNotifier(object):

    def __init__(self, conf, sc):
        self._sc = sc
        self._conf = conf

    def update_status(self, context, notification_data):
        nfp_context = module_context.init()
        resource_data = notification_data['notification'][0]['data']

        request_info = notification_data.get('info')
        request_context = request_info.get('context')
        logging_context = request_context.get('logging_context', {})
        nfp_context['log_context'] = logging_context

        status = resource_data['status']
        LOG.info(_LI("Received VPN's update_status API. "
                     "Making an update_status RPC cast to plugin for object"
                     "with status: %(status)s"),
                 {'status': status})
        rpcClient = transport.RPCClient(a_topics.VPN_NFP_PLUGIN_TOPIC)
        rpcClient.cctxt.cast(context, 'update_status',
                             status=status)

    def ipsec_site_conn_deleted(self, context, notification_data):
        pass


ServicetypeToHandlerMap = {'firewall': FirewallNotifier,
                           'loadbalancer': LoadbalancerNotifier,
                           'loadbalancerv2': LoadbalancerV2Notifier,
                           'vpn': VpnNotifier}


class NaasNotificationHandler(object):

    def __init__(self, conf, sc):
        self.conf = conf
        self.sc = sc

    def handle_notification(self, context, notification_data):
        try:
            LOG.debug("Handling Notification with Data:%s"
                      % notification_data)
            resource_data = notification_data['notification'][0]['data']
            handler = ServicetypeToHandlerMap[notification_data[
                'info']['service_type']](self.conf, self.sc)
            method = getattr(handler, resource_data['notification_type'])
            # Handle RPC Event
            method(context, notification_data)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            msg = ("Generic exception (%s) while handling message (%s) : %s"
                   % (e,
                      notification_data,
                      traceback.format_exception(exc_type,
                                                 exc_value,
                                                 exc_traceback)))
            LOG.error(msg)



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from aim.aim_lib import nat_strategy
from aim import aim_manager
from aim.api import resource as aim_resource
from aim import context as aim_context
from alembic import util as alembic_util
from neutron.db import address_scope_db as as_db
from neutron.db import model_base
from neutron.db import models_v2
import sqlalchemy as sa

from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import apic_mapper
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import db
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import extension_db


class DefunctAddressScopeExtensionDb(model_base.BASEV2):
    # REVISIT: This DB model class is used only for the
    # apic_aim_persist data migration, after which this table is
    # dropped.

    __tablename__ = 'apic_aim_addr_scope_extensions'

    address_scope_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id', ondelete="CASCADE"),
        primary_key=True)
    vrf_dn = sa.Column(sa.String(1024))


def do_apic_aim_persist_migration(session):
    alembic_util.msg(
        "Starting data migration for apic_aim mechanism driver persistence.")

    db_mixin = db.DbMixin()
    aim = aim_manager.AimManager()
    aim_ctx = aim_context.AimContext(session)
    mapper = apic_mapper.APICNameMapper()

    with session.begin(subtransactions=True):
        # Migrate address scopes.
        scope_dbs = (session.query(as_db.AddressScope).
                     all())
        for scope_db in scope_dbs:
            alembic_util.msg("Migrating address scope: %s" % scope_db)
            vrf = None
            ext_db = (session.query(DefunctAddressScopeExtensionDb).
                      filter_by(address_scope_id=scope_db.id).
                      one_or_none())
            if ext_db:
                # It has a pre-existing VRF.
                vrf = aim_resource.VRF.from_dn(ext_db.vrf_dn)
                # REVISIT: Get VRF to verify it exists?
                vrf_owned = False
            if not vrf:
                # It does not have a pre-existing VRF.
                aname = mapper.address_scope(session, scope_db.id)
                vrfs = aim.find(
                    aim_ctx, aim_resource.VRF,
                    name=aname)
                if vrfs:
                    vrf = vrfs[0]
                    vrf_owned = True
            if vrf:
                db_mixin._add_address_scope_mapping(
                    session, scope_db.id, vrf, vrf_owned)
            else:
                alembic_util.warn(
                    "No AIM VRF found for address scope: %s" % scope_db)

        # Migrate networks.
        net_dbs = (session.query(models_v2.Network).
                   all())
        for net_db in net_dbs:
            alembic_util.msg("Migrating network: %s" % net_db)
            bd = None
            epg = None
            vrf = None
            ext_db = (session.query(extension_db.NetworkExtensionDb).
                      filter_by(network_id=net_db.id).
                      one_or_none())
            if ext_db and ext_db.external_network_dn:
                # Its a managed external network.
                ext_net = aim_resource.ExternalNetwork.from_dn(
                    ext_db.external_network_dn)
                # REVISIT: Get ExternalNetwork to verify it exists?
                l3out = aim_resource.L3Outside(
                    tenant_name=ext_net.tenant_name,
                    name=ext_net.l3out_name)
                if ext_db.nat_type == '':
                    ns_cls = nat_strategy.NoNatStrategy
                elif ext_db.nat_type == 'edge':
                    ns_cls = nat_strategy.EdgeNatStrategy
                else:
                    ns_cls = nat_strategy.DistributedNatStrategy
                ns = ns_cls(aim)
                ns.app_profile_name = 'OpenStack'
                for resource in ns.get_l3outside_resources(aim_ctx, l3out):
                    if isinstance(resource, aim_resource.BridgeDomain):
                        bd = resource
                    elif isinstance(resource, aim_resource.EndpointGroup):
                        epg = resource
                    elif isinstance(resource, aim_resource.VRF):
                        vrf = resource
            if not bd:
                # It must be a normal network.
                aname = mapper.network(session, net_db.id)
                bds = aim.find(
                    aim_ctx, aim_resource.BridgeDomain,
                    name=aname)
                if bds:
                    bd = bds[0]
                epgs = aim.find(
                    aim_ctx, aim_resource.EndpointGroup,
                    name=aname)
                if epgs:
                    epg = epgs[0]
                if bd:
                    vrfs = (
                        aim.find(
                            aim_ctx, aim_resource.VRF,
                            tenant_name=bd.tenant_name,
                            name=bd.vrf_name) or
                        aim.find(
                            aim_ctx, aim_resource.VRF,
                            tenant_name='common',
                            name=bd.vrf_name))
                    if vrfs:
                        vrf = vrfs[0]
            if bd and epg and vrf:
                db_mixin._add_network_mapping(
                    session, net_db.id, bd, epg, vrf)
            elif not net_db.external:
                alembic_util.warn(
                    "AIM BD, EPG or VRF not found for network: %s" % net_db)

    alembic_util.msg(
        "Finished data migration for apic_aim mechanism driver persistence.")



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

""" external_connectivity

Revision ID: f16efdc10a71
"""

# revision identifiers, used by Alembic.
revision = 'f16efdc10a71'
down_revision = 'f4d890a9c126'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_external_segments',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('shared', sa.Boolean),
        sa.Column('ip_version', sa.Integer, nullable=False),
        sa.Column('cidr', sa.String(64), nullable=False),
        sa.Column('port_address_translation', sa.Boolean),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_external_policies',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('shared', sa.Boolean),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_ep_to_prs_providing_associations',
        sa.Column('policy_rule_set_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_rule_set_id'],
                                ['gp_policy_rule_sets.id'],
                                ondelete='CASCADE'),
        sa.Column('external_policy_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_policy_id'],
                                ['gp_external_policies.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('policy_rule_set_id',
                                'external_policy_id'))

    op.create_table(
        'gp_ep_to_prs_consuming_associations',
        sa.Column('policy_rule_set_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_rule_set_id'],
                                ['gp_policy_rule_sets.id'],
                                ondelete='CASCADE'),
        sa.Column('external_policy_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_policy_id'],
                                ['gp_external_policies.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('policy_rule_set_id',
                                'external_policy_id'))

    op.create_table(
        'gp_es_to_ep_associations',
        sa.Column('external_policy_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_policy_id'],
                                ['gp_external_policies.id'],
                                ondelete='CASCADE'),
        sa.Column('external_segment_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_segment_id'],
                                ['gp_external_segments.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('external_policy_id',
                                'external_segment_id'))

    op.create_table(
        'gp_external_routes',
        sa.Column('destination', sa.String(64)),
        sa.Column('nexthop', sa.String(64)),
        sa.Column('external_segment_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_segment_id'],
                                ['gp_external_segments.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('external_segment_id',
                                'destination', 'nexthop'))

    op.create_table(
        'gp_nat_pools',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('shared', sa.Boolean),
        sa.Column('ip_version', sa.Integer, nullable=False),
        sa.Column('ip_pool', sa.String(64), nullable=False),
        sa.Column('external_segment_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_segment_id'],
                                ['gp_external_segments.id']),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_es_to_l3p_associations',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['l3_policy_id'],
                                ['gp_l3_policies.id'],
                                ondelete='CASCADE'),
        sa.Column('external_segment_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['external_segment_id'],
                                ['gp_external_segments.id'],
                                ondelete='CASCADE'),
        sa.Column('allocated_address', sa.String(64), nullable=False,
                  primary_key=True),
        sa.PrimaryKeyConstraint(
            'l3_policy_id', 'external_segment_id', 'allocated_address'),
        sa.UniqueConstraint('external_segment_id', 'allocated_address'),
    )


def downgrade():

    op.drop_table('gp_es_to_l3p_associations')
    op.drop_table('gp_nat_pools')
    op.drop_table('gp_external_routes')
    op.drop_table('gp_es_to_ep_associations')
    op.drop_table('gp_ep_to_prs_consuming_associations')
    op.drop_table('gp_ep_to_prs_providing_associations')
    op.drop_table('gp_external_policies')
    op.drop_table('gp_external_segments')
